import { Axis, AxisView } from "./axis";
export class ContinuousAxisView extends AxisView {
}
ContinuousAxisView.__name__ = "ContinuousAxisView";
export class ContinuousAxis extends Axis {
    constructor(attrs) {
        super(attrs);
    }
}
ContinuousAxis.__name__ = "ContinuousAxis";
//# sourceMappingURL=continuous_axis.js.map